from tree_sitter import Language, Parser,Node,TreeCursor,Tree
import tree_sitter_python as tspython
from lib.custom_class import ElementContext
from lib.utils import format_log
import re
class ASTGenerator():
  def __init__(self) -> None:
    # Init parser
    format_log("Initializing AST Parser...")
    self.PY_LANGUAGE = Language(tspython.language(), "python")

    self.parser = Parser()
    self.parser.set_language(self.PY_LANGUAGE)
    format_log("AST Parser initialized.")
  
  def generate(self,ctx:ElementContext)->tuple[Node|None,TreeCursor|None,str]:
    format_log(f"Generating AST for {ctx.get_base_element_name()} in {ctx.location}...")
    # Tentukan jenisnya
    if ctx.type == 'class':
      return self.parse_class(ctx)
    elif ctx.type == 'function':
      return self.parse_function(ctx)
    elif ctx.type == 'module':
      return self.parse_module(ctx)
    elif ctx.type == 'library':
      return self.parse_library(ctx)
    # Selain itu return None
    return (None,None)
  
  def get_all_function_info(self,source_code:str,location:str)->list[dict[str,str]]:
    classes = []
    tree = self.parser.parse(source_code)
    query = self.PY_LANGUAGE.query("""
      (function_definition
        name: (identifier) @function.name)                             
    """)
    captures = query.captures(tree.root_node)
    for node_list in captures:
      # Caritahu parentnya
      node = node_list[0]
      function_name = source_code[node.start_byte:node.end_byte].decode()
      kongteks = ElementContext(function_name,'function',location,[])
      ast,cursor,file_content = self.parse_function(kongteks)
      # Traverse parent
      while node.type not in ['class_definition','module']:
        node = node.parent
      # Cek tipenya
      if node.type == 'class_definition':
        tmp = node.children[1]
        class_name = source_code[tmp.start_byte:tmp.end_byte].decode()
        classes.append({'name':function_name,'type':'class_method','parent':class_name,'ast':ast,'ctx':kongteks,'cursor':cursor})
      elif node.type == 'module':
        classes.append({'name':function_name,'type':'module_function','ast':ast,'ctx':kongteks,'cursor':cursor})
    return classes
  
  def get_dependency_list(self,source_code:str,source_code_method_list:list):
      # Dapatkan daftar dependency beserta lokasi barisnya
      deps_list = []
      for i,line in enumerate(source_code.decode().split("\n")):
        if re.search(r"(from +.+ +import +.+)|(import +.+)",line):
          # Dapetin nama modul
          parts = line.strip().split(" ")[1:] # skip import or from statement
          module_name = {"original":parts[0],"rename":None}
          imported_elmts:list[dict] = []
          i = 1
          while i<len(parts): 
            if(parts[i]==','):
              continue
            elif(parts[i]=='as'):
              i+=1
              if(not imported_elmts):
                # Rename module
                module_name["rename"] = parts[i]
              else:
                # Rename element terakhir
                imported_elmts[-1]["rename"] = parts[i]
            elif(parts[i]=='import'):
              # Import element berikutnya
              i+=1
              if parts[i]=='*':
                for method in source_code_method_list:
                  if(method["type"]=='module_function'):
                    imported_elmts.append({"original":method["name"],"rename":None})
                  elif method["type"]=='class_method':
                    imported_elmts.append({"original":method["parent"],"rename":None})
                break
              else:
                imported_elmts.append({"original":parts[i],"rename":None})
            else:
              imported_elmts.append({"original":parts[i],"rename":None})
            i+=1
          # Konversi imported_elmts
          deps_list.append([i,module_name,imported_elmts])
      return deps_list


  def parse_class(self,ctx:ElementContext)->tuple[Node|None,TreeCursor|None,str]:
    # Buka file
    file_content = ""
    # Generate AST untuk file 
    with open(ctx.location,'rb') as f:
      file_content:bytes = bytes(f.read().decode(),'utf-8')
    tree = self.parser.parse(file_content)
    # Traverse file untuk mencari kelas tersebut
    query = self.PY_LANGUAGE.query("""
      (class_definition 
        name: (identifier) @class.name)                             
    """)
    captures = query.captures(tree.root_node)
    ast = None
    for node in captures:
      if(file_content[node[0].start_byte:node[0].end_byte].decode() == ctx.name.split("(")[0]):
        # Kelasnya ketemu
        ast = node[0].parent
        break
    cursor = tree.walk()
    cursor.reset(ast)
    return (ast,cursor,file_content)

  def parse_function(self,ctx:ElementContext)->tuple[Node|None,TreeCursor|None]:
    # Buka file
    file_content = ""
    # Generate AST untuk file 
    with open(ctx.location,'rb') as f:
      file_content:bytes = bytes(f.read().decode(),'utf-8')
    tree = self.parser.parse(file_content)
    # Traverse file untuk mencari fungsi tersebut
    query = self.PY_LANGUAGE.query("""
      (function_definition
        name: (identifier) @function.name)                             
    """)
    captures = query.captures(tree.root_node)
    ast = None
    for node in captures:
      # get class name
      class_name = ""
      if node[0].parent.parent.type=='block':
        class_name = node[0].parent.parent.parent.children[1].text.decode()
      if(node[0].text.decode() == ctx.name.split("(")[0] or (class_name and f"{class_name}.{node[0].text.decode()}"==ctx.name.split("(")[0])):
        # Fungsinya ketemu
        ast = node[0].parent
        break
    cursor = tree.walk()
    cursor.reset(ast)
    return (ast,cursor,file_content)

  def parse_module(self,ctx:ElementContext)->tuple[Node|None,TreeCursor|None]:
    # Buka file
    file_content = ""
    # Generate AST untuk file 
    with open(ctx.location,'rb') as f:
      file_content:bytes = bytes(f.read().decode(),'utf-8')
    ast = self.parser.parse(file_content).root_node
    cursor = ast.walk()
    return (ast,cursor,file_content)

  def parse_library(self,ctx:ElementContext)->tuple[Node|None,TreeCursor|None]:
    # Tentukan jenis import nya
    try:
      import_type,import_name = ctx.name.split(" ")[:2]
      if import_type == 'from':
        # Import from statement
        # Buka file
        file_content = ""
        # Generate AST untuk file 
        with open(ctx.location,'rb') as f:
          file_content:bytes = bytes(f.read().decode(),'utf-8')
        tree = self.parser.parse(file_content)
        # Traverse file untuk mencari fungsi tersebut
        query = self.PY_LANGUAGE.query("""
          (import_from_statement
            module_name: (dotted_name) @library.name)                             
        """)
        captures = query.captures(tree.root_node)
        ast = None
        for node in captures:
          if(file_content[node[0].start_byte:node[0].end_byte].decode() == import_name.split(" ")[0]):
            # Fungsinya ketemu
            ast = node[0].parent
            break
        cursor = tree.walk()
        cursor.reset(ast)
        return (ast,cursor,file_content)
      elif import_type == 'import':
        # Normal import statement
        # Buka file
        file_content = ""
        # Generate AST untuk file 
        with open(ctx.location,'rb') as f:
          file_content:bytes = bytes(f.read().decode(),'utf-8')
        tree = self.parser.parse(file_content)
        # Traverse file untuk mencari fungsi tersebut
        query = self.PY_LANGUAGE.query("""
          (import_statement
            name: (dotted_name) @library.name)                             
        """)
        captures = query.captures(tree.root_node)
        ast = None
        for node in captures:
          if(file_content[node[0].start_byte:node[0].end_byte].decode() == import_name.strip()):
            # Fungsinya ketemu
            ast = node[0].parent
            break
        cursor = tree.walk()
        cursor.reset(ast)
        return (ast,cursor,file_content)
      else:
        return (None,None,None)
    except ValueError as e:
      return (None,None,None)