from lib.custom_class import ElementContext
from lib.ASTGenerator import ASTGenerator
from lib.utils import format_log
from tree_sitter import Node,TreeCursor
from typing import Optional
from .DependencyManager import FileInformation
from .Const import PATH_SEPARATOR

class CFGGenerator():
  def __init__(self):
    format_log("Initializing CFG Generator...")
    self.ast_parser:ASTGenerator = ASTGenerator()
    format_log("CFG Generator initialized.")

  def generate(self,ctx:ElementContext):
    format_log(f"Generating CFG for {ctx.get_base_element_name()} in {ctx.location}...")
    ast,cursor,source_code = self.ast_parser.generate(ctx)
    method_list = self.ast_parser.get_all_function_info(source_code,ctx.location)
    deps_list = self.ast_parser.get_dependency_list(source_code,method_list)
    dependency = FileInformation(ctx.location.split(PATH_SEPARATOR)[-1],ctx.location,deps_list,method_list)
    return CFG(ast,ctx,cursor,source_code),dependency


class CFG():
  def __init__(self,ast:Node,ctx:ElementContext,cursor:TreeCursor,source_code:str|bytes = "",next_branch:Optional[TreeCursor] = None) -> None:
    self.ast:Node = ast
    self.ctx:ElementContext = ctx
    self.cursor:TreeCursor = cursor
    if isinstance(source_code,str):
      self.source_code = source_code.encode()
    else:
      self.source_code:bytes = source_code
    self.next_branch:Optional[TreeCursor] = next_branch
    self.original_cursor:TreeCursor = self.cursor.copy()
    self.is_traversing:bool = False

  def get_line_index(self,text:str)->int:
    parts = text.split(".")
    # Pastikan tiap bagiannya terpanggi
    for i,line in enumerate(self.source_code.decode().split("\n")):
      if parts[0] in line:
        parts.pop(0)
      if not parts:
        # Dah habis
        return i
    return -1

  def traverse(self)->Node|None:
  # Tentukan jenisnya
    if self.ctx.type == 'class':
      return self.traverse_class()
    elif self.ctx.type == 'function':
      node =  self.traverse_function()
      return node
    elif self.ctx.type == 'module':
      return self.traverse_module()
    elif self.ctx.type == 'library':
      return self.traverse_library()
  
  def reset(self)->None:
    self.cursor = self.original_cursor.copy()
    self.is_traversing = False

  def get_name(self,node:Node)->str:
    if not node:
      return ""
    return self.source_code[node.start_byte:node.end_byte].decode()
  
  def traverse_class(self)->Node|None:
    # Periksa seluruh kelas, cek fungsi
    # Akses body
    self.cursor.goto_last_child()
    have_sibling = True
    while(have_sibling):
      if(self.cursor.node.type=='function_definition'):
        # Cek apakah ada di blacklist konteks
        if(not self.ctx.context or f"Exempt::{self.source_code[self.cursor.node.start_byte:self.cursor.node.end_byte].decode()}"not in self.ctx.context):
          return self.cursor.node
        # Kalau mau analisis kelas selanjutnya, panggil traverse_class/traverse_function
      have_sibling = self.cursor.goto_next_sibling()
    return None

  def traverse_function(self)->Node|None:
    # Terminate
    if self.is_node_equal(self.cursor.node,self.original_cursor.node) and self.is_traversing:
      return None
    if self.is_node_equal(self.cursor.node,self.original_cursor.node):
      self.is_traversing = True

    if(self.cursor.node.type=='function_definition'):
      # Simpan next branch
      tmp_ptr = self.cursor.copy()
      tmp_ptr.goto_next_sibling()
      self.next_branch = tmp_ptr.copy()
      # Geser ke kanan
      node = self.cursor.node
      self.cursor.goto_last_child()
      return node
    elif(self.cursor.node.type=='block'):
      # Blok fungsi
      self.cursor.goto_first_child()
      return self.cursor.node
    elif(self.cursor.node.type=='try_statement'):
      # Blok try-catch
      self.cursor.goto_first_child()
      while(self.cursor.node and self.cursor.node.type not in ['block','except_clause','finally_clause']):
        self.cursor.goto_next_sibling()
      if(self.cursor.node):
        # Branching
        node = self.cursor.node
        self.cursor.goto_next_sibling()
        self.next_branch = self.cursor.copy()
        return node
      else:
        # Coba akses branch berikutnya
        self.cursor.reset_to(self.next_branch)
        return self.cursor.node
    elif(self.cursor.node.type in ['except_clause','finally_clause']):
        self.cursor.goto_last_child()
        return self.cursor.node
    elif(self.cursor.node.type=='if_statement'):
      # Ada percabangan
      # cari block node
      self.cursor.goto_first_child()
      while self.cursor.node.type !='block':
        self.cursor.goto_next_sibling()
      node = self.cursor.node
      # simpan branch berikutnya
      tmp_ptr = self.cursor.copy()
      has_sibling = True
      while has_sibling and tmp_ptr.node.type not in ['elif_clause','else_clause']:
        has_sibling = tmp_ptr.goto_next_sibling()
      if(not has_sibling):
        while not has_sibling:
          prev_node = tmp_ptr.node
          tmp_ptr.goto_parent()
          has_sibling = tmp_ptr.goto_next_sibling()
          if CFG.is_node_equal(prev_node,tmp_ptr.node):
            break
        self.next_branch = tmp_ptr.copy()
      else:
        self.next_branch = tmp_ptr.copy()
      return node
    elif(self.cursor.node.type=='elif_clause'):
      # Ada percabangan
      # cari block node
      self.cursor.goto_first_child()
      while self.cursor.node.type != 'block':
        self.cursor.goto_next_sibling()
      node = self.cursor.node
      self.cursor.goto_parent()
      # simpan branch berikutnya
      while self.cursor.node.type not in ['else_clause']:
        self.cursor.goto_next_sibling()
      self.next_branch = self.cursor.copy()
      return node
    elif(self.cursor.node.type=='else_clause'):
      # Ada percabangan
      # cari block node
      # simpan branch berikutnya
      tmp_ptr = self.cursor.copy()
      self.cursor.goto_parent()
      has_sibling = self.cursor.goto_next_sibling()
      while not has_sibling:
        prev_node = self.cursor.node
        self.cursor.goto_parent()
        has_sibling = self.cursor.goto_next_sibling()
        if CFG.is_node_equal(prev_node,self.cursor.node):
          break
      self.next_branch = self.cursor.copy()
      self.cursor.reset_to(tmp_ptr)
      self.cursor.goto_first_child()
      while self.cursor.node.type != 'block':
        self.cursor.goto_next_sibling()
      node = self.cursor.node
      return node
    elif(self.cursor.node.type in ["with_statement","for_statement","while_statement"]):
      tmp_ptr = self.cursor.copy()
      self.cursor.goto_next_sibling()
      self.next_branch = self.cursor.copy()
      self.cursor.reset_to(tmp_ptr)
      self.cursor.goto_last_child()
      return self.cursor.node
    elif self.cursor.node.type in ['expression_statement']:
      tmp_ptr = self.cursor.copy()
      if not self.cursor.goto_next_sibling():
        while True:
          self.cursor.goto_parent()
          if(self.cursor.node.type!='block'):
            has_sibling = self.cursor.goto_next_sibling()
            if not has_sibling:
              if self.next_branch:
                self.cursor.reset_to(self.next_branch)
                return self.cursor.node
              else:
                self.cursor.reset_to(self.original_cursor)
                return self.cursor.node
            # Skip blok
            else:
              while self.cursor.node.type in ['comment']:
                self.cursor.goto_next_sibling()
            break
      self.next_branch = self.cursor.copy()
      self.cursor.reset_to(tmp_ptr)
      self.cursor.goto_last_child()
      return self.cursor.node

    elif self.cursor.node.type in ['assignment']:
      self.cursor.reset_to(self.next_branch)
      # reset next branch
      self.next_branch = None 
      return self.cursor.node
    elif self.cursor.node.type=='return_statement':
      # Dapatkan yang di return statement
      self.cursor.goto_first_child()
      while self.cursor.node.type=='return':
        self.cursor.goto_next_sibling()
      return self.cursor.node
    elif self.cursor.node.type=='match_statement':
      # Ada percabangan
      # cari block node
      self.cursor.goto_first_child()
      while self.cursor.node.type in ['match',':']:
        self.cursor.goto_next_sibling()
      node = self.cursor.node
      return node
    else:
      # Expression statement dsb
      has_sibling =  self.cursor.goto_next_sibling()
      while not has_sibling and self.cursor.node.type in ["match",":"]:
        has_sibling = self.cursor.goto_next_sibling()
      if(not has_sibling):
        # ke branch berikutnya
        if not self.next_branch:
          # Dah habis
          return None
        self.cursor.reset_to(self.next_branch)
        # reset next branch
        self.next_branch = None
        return self.cursor.node
      else:
        return self.cursor.node

  def traverse_module(self)->Node|None:
    # Periksa seluruh modul, cek kelas sama fungsi
    # Ini isi submodul, isinya kelas atau library
    if(self.cursor.node.type=='decorated_definition'):
      # Pergi ke anaknya
      self.cursor.goto_first_child()
      node = self.cursor.node
      return node 
    elif(self.cursor.node.type=='decorator'):
      self.cursor.goto_first_child()
      node = self.cursor.node
      self.cursor.goto_parent()
      self.cursor.goto_next_sibling()
      return node
    self.cursor.goto_first_child()
    have_sibling = True
    while(have_sibling):
      if(self.cursor.node.type in ['class_definition','function_definition']):
        # Cek apakah ada di blacklist konteks
        if(not self.ctx.context or f"Exempt::{self.source_code[self.cursor.node.start_byte:self.cursor.node.end_byte].decode()}"not in self.ctx.context):
          return self.cursor.node
        # Kalau mau analisis kelas selanjutnya, panggil traverse_class/traverse_function
      have_sibling = self.cursor.goto_next_sibling()
    return None
    

  def traverse_library(self)->Node|None:
    # Tentukan jenisnya
    if(self.cursor.node.type=="import_statement"):
      # Tentukan jenis import nya, apakah dotted_name atau aliases import
      self.cursor.goto_first_child()
      self.cursor.goto_next_sibling()
      if(self.cursor.node.type=="dotted_name"):
        # Format: import a
        self.cursor.goto_first_child()
        # node berisi nama importnya
        return self.cursor.node
      elif(self.cursor.node.type=="aliased_import"):
        # Format: import a as b
        self.cursor.goto_first_child()
        self.cursor.goto_last_child()
        # node berisi nama importnya
        return self.cursor.node
    elif(self.cursor.node.type=="import_from_statement"):
        # Format: from a import c
        for i in range(len(self.cursor.node.children)-1,-1,-1):
          if self.cursor.node.children[i].type=="wildcard_import":
            # Format: from a import *
            self.cursor.goto_first_child()
            self.cursor.goto_next_sibling()
            return self.cursor.node # return a
          elif(self.cursor.node.children[i].type=="dotted_name"):
            self.cursor.goto_last_child()
            while self.cursor.node.type in [')',',','comment']:
              self.cursor.goto_previous_sibling()
            for _ in range(len(self.cursor.node.children)-i):
              self.cursor.goto_previous_sibling()
            node = self.cursor.node
            return node
          elif(self.cursor.node.children[i].type=="aliased_import"):
            # format: from a import b as c
            self.cursor.goto_last_child()
            for _ in range(len(self.cursor.node.children)-i):
              self.cursor.goto_previous_sibling()
            self.cursor.goto_first_child()
            self.cursor.goto_last_child()
            node = self.cursor.node
            self.cursor.goto_parent()
            self.cursor.goto_parent()
            return node
        # node berisi nama importnya
        return self.cursor.node
    elif(self.cursor.node.type in ["dotted_name","aliased_import"]):
        # Format: from a import, [b],c,d, [e as f]
        self.cursor.goto_previous_sibling()
        while(True):
          if(self.cursor.node.type=="import"):
            # Udah habis listnya
            self.cursor = None
            return None
          # cek tipenya
          if(self.cursor.node.type=='dotted_name'):
            # Cek apakah ada di list konteks
            if(not self.ctx.context or self.source_code[self.cursor.node.start_byte:self.cursor.node.end_byte].decode() in self.ctx.context):
              return self.cursor.node
          elif(self.cursor.node.type=='aliased_import'):
            # Cek apakah ada di list konteks
            self.cursor.goto_first_child()
            self.cursor.goto_last_child()
            node = self.cursor.node
            self.cursor.goto_parent()
            self.cursor.goto_parent()
            if(not self.ctx.context or self.source_code[node.start_byte:node.end_byte].decode() in self.ctx.context):
              return node
          self.cursor.goto_previous_sibling()        
    return None
  @staticmethod
  def is_node_equal(node:Node,expected:Node):
    return node and expected and node.byte_range == expected.byte_range